﻿package stego
{
    import flash.display.Sprite;
    import flash.events.Event;
    import flash.events.MouseEvent;
    import flash.geom.Point;
    import flash.geom.Rectangle;
    import stego.GameInterface;
    import stego.GameObject;
    import stego.debug.DebugWidget;
    import stego.debug.Profiler;
    import stego.util.InputUtil;
    import stego.util.MathUtil;
    import stego.util.PhysicsUtil;
    import stego.ui.PauseScreen;
    import stego.GameState;
    import stego.util.CheatUtil;
    import stego.util.music.MusicUtil;
    import stego.util.scene.GameObjectHolder;
    import stego.util.scene.Scene;
    import stego.util.TimeUtil;
    import stego.ui.MusicCredits;

    /**
     * Base class that you will be extending to create your own games.
     * @author Mike Meyer
     */
    public class Game extends Sprite implements GameInterface
    {
        protected var gameState:GameStateInterface;
        protected var nextGameState:GameStateInterface;
        protected var isPaused:Boolean;

        /// widget for music credits
        protected var musicCredits:MusicCredits;

        protected var _cameraPosition:Point = new Point(0, 0);
        public function get cameraX():Number { return _cameraPosition.x; }
        public function set cameraX(value:Number):void { _cameraPosition.x = value; cameraUpdate(); }
        public function get cameraY():Number { return _cameraPosition.y; }
        public function set cameraY(value:Number):void { _cameraPosition.y = value; cameraUpdate(); }

        /// see how much the game is zoomed in
        public function get zoom():Number { return gameHolder.scaleX; }

        /// set how much the game is zoomed in
        public function set zoom(zoom:Number):void { gameHolder.scaleX = zoom; gameHolder.scaleY = zoom; }

        protected var pauseScreen:GameObject;
        protected var scene:GameObjectHolder;
        protected var gameHolder:GameObject;

        protected var _cheatUtil:CheatUtil;
        public function get cheats():CheatUtil { return _cheatUtil; }

        protected var _musicUtil:MusicUtil;
        public function get music():MusicUtil { return _musicUtil; }

        protected var _layers:Scene;
        public function get layers():Scene { return _layers; }

        protected var _time:TimeUtil;
        public function get time():TimeUtil { return _time; }

        protected var _physics:PhysicsUtil;
        public function get physics():PhysicsUtil { return _physics; }
        public function set physics(value:PhysicsUtil):void { _physics = value; }

        protected var _debugger:DebugWidget;
        public function get debug():DebugWidget { return _debugger; }

        protected var _cameraShake:Number = 0;
        public function get cameraShake():Number { return _cameraShake; }
        public function set cameraShake(value:Number):void { _cameraShake = value;  }

        protected var _cameraFollowTarget:GameObject = null;
        public function get cameraFollowTarget():GameObject { return _cameraFollowTarget; }
        public function set cameraFollowTarget(value:GameObject):void { _cameraFollowTarget = value;  }

        protected var _cameraFollowRect:Rectangle = null;
        public function get cameraFollowRect():Rectangle { return _cameraFollowRect; }
        public function set cameraFollowRect(value:Rectangle):void { _cameraFollowRect = value;  }

        protected var _cameraSpeed:Number = 1000;
        public function get cameraSpeed():Number { return _cameraSpeed; }
        public function set cameraSpeed(value:Number):void { _cameraSpeed = value; }

        protected var _inputUtil:InputUtil;
        public function get inputUtil():InputUtil { return _inputUtil; }
        public function set inputUtil(value:InputUtil):void { _inputUtil = value; }

        public function get stageRect():Rectangle { return new Rectangle(0, 0, stageWidth, stageHeight); }

        /** whether debug stuff should be shown */
        public function get debuggingFeaturesAllowed():Boolean { return _debuggingFeaturesAllowed; }
        public function set debuggingFeaturesAllowed(value:Boolean):void
        {
            _debuggingFeaturesAllowed = value;
            if (value)
            {
                if (!_debugger && scene) { createDebugger(); }
            }
            else
            {
                if (_debugger) { _debugger.visible = false; }
            }
        }
        protected var _debuggingFeaturesAllowed:Boolean;

        // shake looks much better if the shaking tends to
        // go back and forth instead of actually random
        protected var _cameraShakeDir:Number = 0;
        protected var minShakeRotation:Number = 0.67 * Math.PI;
        protected var maxShakeRotation:Number = 1.33 * Math.PI;

        protected var _stageWidth:Number;
        public function get stageWidth():Number { return _stageWidth; }
        protected var _stageHeight:Number;
        public function get stageHeight():Number { return _stageHeight; }

        /**
         * Automagically called by the preloader for the initial init,
         * but if you shut it down and want to restart, you'll need to call it yourself.
         *
         * override (but don't forget to call super.init() to customize your game
         */
        public function init(stageWidth:Number, stageHeight:Number):void
        {
            // allow debug stuff only when running locally
            // letting the setter do most of the setup for this widget
            debuggingFeaturesAllowed = (loaderInfo.url.indexOf("file") == 0);

            if (null == stage)
            {
                throw new Error("Called init before adding game to the stage!  Please add it first!");
            }

            _stageWidth = stageWidth;
            _stageHeight = stageHeight;

            _cameraPosition = new Point(
                0.5 * _stageWidth,
                0.5 * _stageHeight );

            _cheatUtil = new CheatUtil();
            cheats.init(stage);

            _musicUtil = new MusicUtil();
            music.init();

            _time = new TimeUtil();
            time.init();

            _physics = new PhysicsUtil();
            _physics.init(this);

            cameraFollowTarget = null;

            isPaused = false;

            _inputUtil = new InputUtil(stage);

            nextGameState = null;

            scene = new GameObjectHolder(this, this);
            setupScene();

            // pause/unpause need to be handled AFTER we set up the pause screen
            addEventListener(Event.ENTER_FRAME, handleEnterFrame);
            stage.addEventListener(Event.DEACTIVATE, _pause);
            stage.addEventListener(MouseEvent.CLICK, _unpause);

            // hit tilde to toggle the debug layer
            cheats.addCheatCode('`', toggleDebug);
        }

        protected function toggleDebug():void
        {
            if (_debugger) { _debugger.visible = !_debugger.visible; }
        }

        /**
         * add the persistent display objects that stick around even between levels
         */
        protected function setupScene():void
        {
            createSceneLayers();
            createMusicCredits();
            createPauseIndicator();

            if (_debuggingFeaturesAllowed)
            {
                createDebugger();
            }
        }

        /**
         * create the layers util
         */
        protected function createSceneLayers():void
        {
            // holder for the game itself (gets moved to scroll the camera)
            gameHolder = scene.add();

            _layers = new Scene();
            layers.init(gameHolder, this);
        }

        /**
         * create the debugger widget
         */
        protected function createDebugger():void
        {
            _debugger = scene.add(DebugWidget) as DebugWidget;
        }

        /**
         * create the music credits widget
         */
        protected function createMusicCredits():void
        {
            musicCredits = scene.add(MusicCredits) as MusicCredits;
            musicCredits.musicUtil = music;
        }

        /**
         * move the music credits widget to the bottom on the screen
         */
        protected function moveMusicCreditsToBottom():void
        {
            musicCredits.y = stageHeight - musicCredits.height;
        }

        /**
         * create the pause screen
         */
        protected function createPauseIndicator():void
        {
            pauseScreen = scene.add(stego.ui.PauseScreen);
            pauseScreen.visible = false;
        }

        protected function handleEnterFrame(e:Event = null):void
        {
            time.timeDelta = (1.0 / stage.frameRate) * time.timeScale;

            doGameStateChange();

            if (!isPaused)
            {
                update();
                integrate();
                preCollision();
                collision();
                preDraw();
            }
            inputUtil.update();
        }

        protected function update():void
        {
            layers.update(time.timeDelta);
            scene.update(time.timeDelta);
            if (null != gameState)
            {
                gameState.update();
            }
            physics.update(time.timeDelta);

            // camera shake
            _cameraShakeDir += MathUtil.lerp( minShakeRotation, maxShakeRotation, Math.random() );
            _cameraShakeDir = MathUtil.wrap( _cameraShakeDir, 0, 2.0 * Math.PI);
            cameraUpdate();
        }

        /**
         * update the movieclips to reflect the new camera position
         */
        protected function cameraUpdate():void
        {
            if (null != cameraFollowTarget)
            {
                var newPos:Point = MathUtil.moveToward2d(_cameraPosition, new Point(cameraFollowTarget.x, cameraFollowTarget.y), cameraSpeed * time.timeDelta);

                if (null != cameraFollowRect)
                {
                    _cameraPosition.x = MathUtil.clamp(newPos.x, cameraFollowRect.left, cameraFollowRect.right);
                    _cameraPosition.y = MathUtil.clamp(newPos.y, cameraFollowRect.top, cameraFollowRect.bottom);
                }
            }

            // TODO: there probably should be some localToGlobal tricks in
            // here in case we do crazier transforms than just zoom
            gameHolder.x = (0.5 * stageWidth) - (_cameraPosition.x * zoom) + (cameraShake * Math.cos(_cameraShakeDir));
            gameHolder.y = (0.5 * stageHeight) - (_cameraPosition.y * zoom) + (cameraShake * Math.sin(_cameraShakeDir));
        }

        protected function integrate():void
        {
            layers.integrate(time.timeDelta);
            scene.integrate(time.timeDelta);
            physics.integrate(time.timeDelta);
        }

        protected function preCollision():void
        {
            layers.preCollision(time.timeDelta);
            if (null != gameState)
            {
                gameState.preCollision();
            }
            physics.preCollision(time.timeDelta);
        }
        protected function collision():void
        {
            layers.collision(time.timeDelta);
            if (null != gameState)
            {
                gameState.collision();
            }
            physics.collision(time.timeDelta);
        }

        protected function preDraw():void
        {
            layers.preDraw(time.timeDelta);
            scene.preDraw(time.timeDelta);
            if (null != gameState)
            {
                gameState.preDraw();
            }
            physics.preDraw(time.timeDelta);
        }

        /**
         * freeze time and show the pause screen
         */
        public function pause():void
        {
            isPaused = true;
            pauseScreen.visible = true;
            music.pause();
        }

        /**
         * light wrapper for using Pause as an event callback
         */
        protected function _pause(e:Event):void
        {
            pause();
        }

        /**
         * resume after pausing
         */
        public function unpause():void
        {
            isPaused = false;
            pauseScreen.visible = false;
            music.resume();
        }

        /**
         * light wrapper for using Pause as an event callback
         */
        protected function _unpause(e:Event):void
        {
            if (isPaused)
            {
                unpause();
            }
        }

        /**
         * if a script change was queued up, perform the actual switch now
         */
        protected function doGameStateChange():void
        {
            if (nextGameState != null)
            {
                endScript(gameState);
                var temp:GameStateInterface = nextGameState; // careful! something triggered by Begin might change nextGameState!
                nextGameState = null;

                beginScript(temp);
                gameState = temp;
            }
        }

        /**
         * shut it down
         */
        protected function endScript(script:GameStateInterface):void
        {
            if (script != null)
            {
                script.end();
            }
        }

        /**
         *
         */
        protected function beginScript(script:GameStateInterface):void
        {
            if (script != null)
            {
                script.begin(this);
            }
        }

        /**
         * set a script to switch to once the current update call is finished
         */
        public function changeGameState(nextGameState:GameStateInterface):void
        {
            this.nextGameState = nextGameState;
        }

        /**
         * go back to the starting position
         */
        public function resetScrolling():void
        {
            gameHolder.x = 0;
            gameHolder.y = 0;
        }

        /**
         * Clean up
         */
        public function shutdown():void
        {
            cameraFollowTarget = null;
            gameState = null;
            nextGameState = null;

            removeEventListener(Event.ENTER_FRAME, handleEnterFrame);
            stage.removeEventListener(Event.DEACTIVATE, _pause);
            stage.removeEventListener(MouseEvent.CLICK, _unpause);

            if (null != gameHolder)
            {
                scene.remove(gameHolder);
                gameHolder = null;
            }

            if (null != pauseScreen)
            {
                scene.remove(pauseScreen);
                pauseScreen = null;
            }

            if (null != this.cheats)
            {
                this.cheats.shutdown();
                _cheatUtil = null;
            }

            if (null != _musicUtil)
            {
                _musicUtil.shutdown();
                _musicUtil = null;
            }

            if (null != _layers)
            {
                _layers.shutdown();
                _layers = null;
            }

            if (null != _physics)
            {
                _physics.shutdown();
                _physics = null;
            }

            if (null != _time)
            {
                _time.shutdown();
                _time = null;
            }

            if (null != scene)
            {
                scene.removeAll();
                scene = null;
            }
        }
    }
}
