﻿package stego
{
    import flash.display.DisplayObject;
    import flash.display.MovieClip;
    import flash.display.Sprite;
    import flash.events.Event;
    import flash.geom.Point;
    import flash.geom.Rectangle;
    import stego.debug.Profiler;
    import stego.script.step.AnimateTo;
    import stego.util.ArrayUtil;
    import stego.util.collision.CollisionInfo;
    import stego.util.ObjectPool;
    import stego.util.scene.GameObjectHolder;
    import stego.util.schedule.Schedule;
    import stego.behavior.IBehavior;

    /**
     * ...
     * @author Mike Meyer
     */
    public class GameObject extends MovieClip
    {
        public var dx:Number = 0.0;
        public var dy:Number = 0.0;
        public var dr:Number = 0.0;

        /// workaround for Flash rounding to twips
        private var _x:Number = 0.0;
        public override function get x () : Number { return _x; }
        public override function set x (value:Number) : void { _x = value; super.x = value; }

        /// workaround for Flash rounding to twips
        private var _y:Number = 0.0;
        public override function get y () : Number { return _y; }
        public override function set y (value:Number) : void { _y = value; super.y = value; }

        /// so that physics utilities can tell what they need to ignore
        public var physicsEnabled:Boolean;

        /// so that physics utilities can tell what they need to ignore
        public var affectedByGravity:Boolean;

        public var alive:Boolean;

        private var _behaviors:Array;

        /**
         * whether the object is resting on the ground
         * invalid while inside PhysicsUtil.update, so don't try to use it then.
         */
        public var isOnGround:Boolean;

        // check to see if a game object is on screen
        public function get isOnScreen():Boolean
        {
            var rect:Rectangle = getCollisionBounds();
            return    (rect.left   < stage.stageWidth)
                   && (rect.right  > 0)
                   && (rect.top    < stage.stageHeight)
                   && (rect.bottom > 0) ;
        }

        /** scheduling utility used for animations */
        protected var _scheduler:Schedule;

        /**
         * GameObjectHolder for managing internal gameobjects
         */
        private var _holder:GameObjectHolder;
        public function get internalGameObjects():GameObjectHolder { return _holder; }

        /**
         * initial setup and preparation for reuse
         */
        public function init(game:GameInterface):void
        {
            this.x = 0.0;
            this.y = 0.0;
            this.dx = 0.0;
            this.dy = 0.0;
            this.visible = true;
            physicsEnabled = true;
            affectedByGravity = true;
            if (null == _holder)
            {
                _holder = new GameObjectHolder(this, game);
            }
            alive = true;
        }

        public function integrate(elapsedTime:Number):void
        {
            this.x += dx * elapsedTime;
            this.y += dy * elapsedTime;
            this.rotation += dr * elapsedTime;

            internalGameObjects.integrate(elapsedTime);
        }

        /**
         * per-frame updates
         * @param	elapsedTime - how long the last frame was
         */
        public function update(elapsedTime:Number):void
        {
            internalGameObjects.update(elapsedTime);

            if (null != _behaviors)
            {
                for each (var b:IBehavior in _behaviors)
                {
                    b.update(elapsedTime);
                }
            }

            if (_scheduler)
            {
                _scheduler.update(elapsedTime);
            }
        }

        /**
         * called after all the final positions have been figured
         * just before drawing
         */
        public function preDraw(timeDelta:Number):void
        {
            internalGameObjects.preDraw(timeDelta);
        }

        /**
         * Get the collision bounds for collision testing
         * if ret is provided, the result will be stored there (and returned)
         * otherwise it will create a new rect
         */
        public function getCollisionBounds(ret:Rectangle = null):Rectangle
        {
            if (!ret) { ret = new Rectangle(); }
            var bounds:Rectangle = this.getBounds(this.parent);

            ret.x = bounds.x;
            ret.y = bounds.y;
            ret.width = bounds.width;
            ret.height = bounds.height;
            return ret;
        }

        /**
         * cleanup for when we are done with this instance for a while.
         * May still get reused!
         */
        public function shutdown():void
        {
            if (_scheduler)
            {
                _scheduler.shutdown();
                ObjectPool.returnObject(Schedule, _scheduler);
                _scheduler = null;
            }
            removeAllBehaviors();
        }

        /**
         * override to deal with collisions
         */
        public function collisionOccurred(ci:CollisionInfo, object:GameObject):void
        {
            if (!canPassThrough(ci, object))
            {
                var collisionBox:Rectangle = this.getCollisionBounds();
                var offsetIntoCollisionBoxFromLeft:Number = x - collisionBox.left;
                var offsetIntoCollisionBoxFromTop:Number = y - collisionBox.top;
                var offsetIntoCollisionBoxFromRight:Number = x - collisionBox.right;
                var offsetIntoCollisionBoxFromBottom:Number = y - collisionBox.bottom;

                if ( ci.normal.x > 0.0 )
                {
                    x = ci.location.x + offsetIntoCollisionBoxFromLeft;
                    if (dx < 0.0) { dx = 0.0; }
                }
                else if ( ci.normal.x < 0.0 )
                {
                    x = ci.location.x + offsetIntoCollisionBoxFromRight;
                    if (dx > 0.0) { dx = 0.0; }
                }

                if ( ci.normal.y > 0.0  )
                {
                    y = ci.location.y + offsetIntoCollisionBoxFromTop;
                    if (dy < 0.0) { dy = 0.0; }
                }
                else if ( ci.normal.y < 0.0 )
                {
                    y = ci.location.y + offsetIntoCollisionBoxFromBottom;
                    if (dy > 0.0) { dy = 0.0; }
                }

                isOnGround = isOnGround || (ci.normal.y < -Math.SQRT1_2);
            }
        }

        /**
         * override to let yourself go through things
         */
        public function canPassThrough(ci:CollisionInfo, object:GameObject):Boolean
        {
            return false;
        }

        /**
         * start an animation
         * @param	startValues - where it should start at.  example: {x:100}
         * @param	endValues - where it is animating to.  example: {x:200}
         * @param	animDuration - how long it takes to run the animation
         */
        public function animate(startValues:Object, endValues:Object, animDuration:Number=1.0):void
        {
            if (!_scheduler) { _scheduler = ObjectPool.getObject(Schedule); }
            _scheduler.runScript(new stego.script.step.Animate(this, startValues, endValues, animDuration));
        }

        /**
         * start an animation
         * @param	objectToAnimate - what we're animating
         * @param	endValues - where it is animating to.  example: {x:200}
         * @param	animDuration - how long it takes to run the animation
         */
        public function animateTo(endValues:Object, animDuration:Number=1.0):void
        {
            if (!_scheduler) { _scheduler = ObjectPool.getObject(Schedule); }
            _scheduler.runScript(new stego.script.step.AnimateTo(this, endValues, animDuration));
        }

        public function addBehavior(value:IBehavior):void
        {
            if (null == _behaviors)
            {
                _behaviors = new Array();
            }
            _behaviors.push(value);
        }
        public function removeBehavior(value:IBehavior):void
        {
            ArrayUtil.removeEqual(_behaviors, value);
        }
        public function removeAllBehaviors():void
        {
            if (_behaviors) { _behaviors.length = 0; }
        }

        /**
         * Setup that needs to happen before we do the collision checks
         */
        public function preCollision(timeDelta:Number):void
        {
            isOnGround = false;

            internalGameObjects.preCollision(timeDelta);
        }

        /**
         * do collision checks on internal stuff
         */
        public function collision(timeDelta:Number):void
        {
            internalGameObjects.preCollision(timeDelta);
        }

        /**
         * check whether this game object is overlapping another
         */
        public function isOverlapping(go:GameObject):Boolean
        {
            return getCollisionBounds().intersects(go.getCollisionBounds());
        }
    }
}
