﻿package stego
{
    import flash.display.DisplayObject;
    import flash.display.MovieClip;
    import flash.display.Sprite;
    import flash.events.Event;
    import flash.utils.getDefinitionByName;
    import stego.GameInterface;
    import stego.GameStateInterface;

    /**
     * Base preloader. Extend and set as your startup class.
     *
     * game-specific implementations probably need nothing more than
     * public function Preloader()
     * {
     *     super(<width>, <height>, <name of "Main" class>);
     * }
     */
    public class Preloader extends Sprite
    {
        protected var loadingBarWidth:Number = 300;
        protected var loadingBarHeight:Number = 30;
        protected var loadingBarBorderWidth:Number = 2;
        protected var loadingBarMargin:Number = 3;
        protected var loadingBarColor:Number = 0xFFFFFF;
        protected var loadingBarBackgroundColor:Number = 0x000000;
        protected var useBackground:Boolean = false;

        // pixel distance to "shake" to
        protected var shakeMagnitude:Number = 5;

        /// max loading bar progress per frame.
        /// change temporarily to see what the loading bar looks like
        private static const MAX_PROGRESS_PER_FRAME:Number = 1.0;

        // hard to name... values <1 decrease how often the shakes happen
        private static const RANDOMNESS_SCALE:Number = 0.5;

        private var shakeDirection:Number = 0;

        private var loadingBackground:Sprite;
        private var loadingBarBorder:Sprite;
        private var loadingBarProgress:Sprite;

        // working around an apparent Flash bug
        private var previousProgress:Number;

        protected var stageWidth:Number;
        protected var stageHeight:Number;
        protected var className:String;
        protected var initialGameStateClassName:String;

        /**
         * constructor
         * @param	stageWidth  - width of the game
         * @param	stageHeight - height of the game
         * @param	className   - name of the "main" game class to instantiate once everything has been preloaded
         * @param	initialGameStateClassName - name of a gamestate to amke the game switch to right away.
         */
        public function Preloader( stageWidth:Number = 480,
                                   stageHeight:Number = 320,
                                   className:String = "stego.core.BaseGame",
                                   initialGameStateClassName:String = null )
        {
            this.stageWidth = stageWidth;
            this.stageHeight = stageHeight;
            this.className = className;
            this.initialGameStateClassName = initialGameStateClassName;

            addEventListener(Event.ENTER_FRAME, checkProgress);

            previousProgress = 0;

            // show loader
            createGraphics();
        }

        /**
         * called every frame to update the progress bar and check for when loading is complete
         */
        private function checkProgress(e:Event):void
        {
            var p:Number = (this.loaderInfo.bytesLoaded / this.loaderInfo.bytesTotal);
            // hack to make testing the loading bar easier
            var cappedProgess:Number = previousProgress + MAX_PROGRESS_PER_FRAME;
            if (p > cappedProgess)
            {
                p = cappedProgess;
            }

            // vibrate as it fills
            if (Math.random() < p*p*RANDOMNESS_SCALE)
            {
                shakeDirection = shakeDirection + (Math.PI * 0.33) * (Math.random() + 1);
                if (shakeDirection > 2 * Math.PI)
                {
                    shakeDirection -= 2 * Math.PI;
                }
                x = Math.cos(shakeDirection) * p * shakeMagnitude;
                y = Math.sin(shakeDirection) * p * shakeMagnitude;
            }
            else
            {
                // drop back when not jerking around
                x *= 0.9;
                y *= 0.9;
            }

            // if p seems to have gotten invalid, just skip the preloader stuff
            if ( (previousProgress > p) || (p >= 1.0) )
            {
                removeEventListener(Event.ENTER_FRAME, checkProgress);

                // make sure we undo the shaking
                x = 0;
                y = 0;

                startup();
            }
            else
            {
                previousProgress = p;

                loadingBarProgress.scaleX = p;
            }
        }

        /**
         * once preloading is complete, start the game up
         */
        private function startup():void
        {
            // hide loader
            removeGraphics();

            // mask out anything that would go outside of the game area
            var mask:Sprite = new Sprite();
            mask.graphics.beginFill(0xFF00FF);
            mask.graphics.drawRect(
                0,
                0,
                stageWidth,
                stageHeight );
            mask.graphics.endFill();
            addChild(mask);

            var mainClass:Class = getDefinitionByName(className) as Class;
            var mainAsDo:DisplayObject = new mainClass();
            var mainAsGame:GameInterface = mainAsDo as GameInterface;
            addChild(mainAsDo);
            mainAsGame.init(stageWidth, stageHeight);
            mainAsDo.mask = mask;

            // go to the game state if they provided one
            if (null != initialGameStateClassName)
            {
                var stateClass:Class = getDefinitionByName(initialGameStateClassName) as Class;
                var state:GameStateInterface = new stateClass();
                mainAsGame.changeGameState(state);
            }
        }

        protected function createGraphics():void
        {
            var barX:Number = 0.5 * ( stageWidth - loadingBarWidth);
            var barY:Number = 0.5 * ( stageHeight - loadingBarHeight);

            if (useBackground)
            {
                loadingBackground = new Sprite();
                loadingBackground.graphics.beginFill(loadingBarBackgroundColor);
                loadingBackground.graphics.drawRect(
                    -shakeMagnitude,
                    -shakeMagnitude,
                    stageWidth + 2 * shakeMagnitude,
                    stageHeight + 2 * shakeMagnitude ); // making it "too" big to account for screen shake
                loadingBackground.graphics.endFill();
                this.addChild(loadingBackground);
            }

            loadingBarBorder = new Sprite();
            loadingBarBorder.graphics.lineStyle(loadingBarBorderWidth, loadingBarColor, 1);
            loadingBarBorder.graphics.beginFill(loadingBarColor, 0.0);
            loadingBarBorder.graphics.drawRect(
                0,
                0,
                loadingBarWidth,
                loadingBarHeight);
            loadingBarBorder.graphics.endFill();
            loadingBarBorder.x = barX;
            loadingBarBorder.y = barY;
            this.addChild(loadingBarBorder);

            loadingBarProgress = new Sprite();
            loadingBarProgress.graphics.lineStyle(loadingBarBorderWidth, loadingBarColor, 0);
            loadingBarProgress.graphics.beginFill(loadingBarColor);
            loadingBarProgress.graphics.drawRect(
                0,
                0,
                loadingBarWidth - (2.0 * loadingBarMargin),
                loadingBarHeight - (2.0 * loadingBarMargin) );
            loadingBarProgress.graphics.endFill();
            loadingBarProgress.x = barX + loadingBarMargin;
            loadingBarProgress.y = barY + loadingBarMargin;
            this.addChild(loadingBarProgress);
        }

        private function removeGraphics():void
        {
            if (useBackground)
            {
                this.removeChild(loadingBackground);
                loadingBackground = null;
            }
            this.removeChild(loadingBarBorder);
            loadingBarBorder = null;
            this.removeChild(loadingBarProgress);
            loadingBarProgress = null;
        }
    }
}
