package stego.behavior.platformer
{
    import stego.behavior.Base;
    import stego.behavior.IBehavior;
    import stego.constants.Keyboard;
    import stego.util.InputUtil;
    import stego.gameobject.PlatformerCharacter;
    import stego.util.MathUtil;

    /**
     * Behavior for jumping.  Made to go nice with Run.
     * @author Mike Meyer
     */
    public class Jump extends Base
    {
        /** array of keys than you can use to jump */
        public var jumpKeys:Array = [Keyboard.SPACE, Keyboard.SHIFT, Keyboard.UP, Keyboard.Z, Keyboard.X];

        /** remember which key they used to jump, so we know which key's release event we care about */
        protected var _jumpKey:uint = 0;

        /** whether the character is currently in the middle of a jump */
        public var isJumping:Boolean = false;

        /**
         * how many times the character can jump when already in the air
         * can be negative one if you want infinite air jumps
         */
        public var numAirJumps:int = 0;

        /** how many air jumps remain.  gets reset when he hits the ground. */
        protected var _airJumpsRemaining:Number = numAirJumps;

        /** how fast to go upward when he jumps */
        public var jumpSpeed:Number = 297;

        /** how fast to go slow down the upward motion when you let go of the jump button */
        public var jumpStopAccel:Number = 0;

        /** if set, this gets called with no args when the player jumps */
        public var jumpCallback:Function;

        /**
         * constructor
         * @param	character
         * @param	inputUtil
         */
        public function Jump(character:PlatformerCharacter, inputUtil:InputUtil)
        {
            super(character, inputUtil);
        }

        /**
         * process one frame of controls
         */
        override public function update(elapsedTime:Number):void
        {
            if ( inputUtil.isUp(_jumpKey) || character.isOnGround)
            {
                isJumping = false;
                _jumpKey = 0;
            }

            // air jumps always get reset when you touch the ground
            if (character.isOnGround) { resetAirJumps(); }

            // if we aren't holding the jump button anymore, cut
            // off the jump height
            if (!character.isOnGround && !isJumping && (character.dy < 0) )
            {
                character.dy += jumpStopAccel * elapsedTime;

                // don't shoot us down, though
                if (character.dy > 0)
                {
                    character.dy = 0;
                }
            }

            // starting a new jump
            var justPressedJumpKey:uint = 0;
            for each (var keyCode:uint in jumpKeys)
            {
                if (inputUtil.wasJustPressed(keyCode))
                {
                    justPressedJumpKey = keyCode;
                }
            }

            if (inputUtil.wasJustPressed(justPressedJumpKey))
            {
                jumpKeyPressed(justPressedJumpKey);
            }

            // reset air jumps after landing
            if (character.isOnGround)
            {
                _airJumpsRemaining = numAirJumps;
            }
        }

        /**
         * overridable handling of the jump button press
         */
        protected function jumpKeyPressed(justPressedJumpKey:uint):void
        {
            if (character.isOnGround || (_airJumpsRemaining != 0) )
            {
                isJumping = true;
                _jumpKey = justPressedJumpKey;

                character.dy = -jumpSpeed;

                if (null != jumpCallback)
                {
                    jumpCallback();
                }
            }
        }

        /**
         * reset the number of air jumps left to the max
         */
        public function resetAirJumps():void
        {
            _airJumpsRemaining = numAirJumps;
        }
    }
}
