package stego.collections
{
    import flash.utils.Dictionary;
    import stego.util.ObjectPool;
	/**
     * Quick+dirty linked list for faster insertion/deletion than you get with
     * arrays.
     * Untyped-ness may be an issue when accessing members.  I have not actually p[rofiled this yet.
     * @author Mike Meyer
     */
    public class LinkedList
    {
        /**
         * first item in the list (may be null)
         */
        public var first:LinkedListNode;

        /**
         * last item in the list (may be null)
         */
        public var last:LinkedListNode;

        /**
         * how many items are in the list
         */
        public var count:uint;

        /**
         * lookup table for finding nodes by value
         * Set quickLookup to true in the constructor to make it use this.
         * Note that if that's set to true, you can't store the same value
         * in the linked list twice (and it must contain objects).
         */
        private var quickLookUpTable:Dictionary;

        /**
         * constructor
         * @param quickLookUp   Create a mapping from values to nodes for fast
         *                      lookup.  Note that if you use this, then you
         *                      can't store the same value in the linked list
         *                      twice (and it must contain objects).
         */
        public function LinkedList(quickLookUp:Boolean = false)
        {
            first = last = null;
            count = 0;

            if (quickLookUp)
            {
                quickLookUpTable = new Dictionary();
            }
        }

        /**
         * Search for the first node that has the given value.
         * Fast if you constructed this with quickLookUp = true
         * returns null if the value was not found
         * @param	value   value to search for
         */
        public function find(value:*):LinkedListNode
        {
            if (quickLookUpTable)
            {
                return quickLookUpTable[value];
            }
            else
            {
                var node:LinkedListNode = first;
                var next:LinkedListNode;
                while (node)
                {
                    next = node.next;
                    if (node.value == value)
                    {
                        return node;
                    }
                    node = next;
                }
            }

            return null;
        }

        /**
         * Search for the last node that has the given value.
         * Fast if you constructed this with quickLookUp = true
         * returns null if the value was not found
         * @param	value   value to search for
         */
        public function rfind(value:*):LinkedListNode
        {
            if (quickLookUpTable)
            {
                return quickLookUpTable[value];
            }
            else
            {
                var node:LinkedListNode = last;
                var prev:LinkedListNode;
                while (node)
                {
                    prev = node.prev;
                    if (node.value == value)
                    {
                        return node;
                    }
                    node = prev;
                }
            }

            return null;
        }

        /**
         * Remove a node from the list.
         */
        public function remove(node:LinkedListNode):void
        {
            // early out on removing null
            if (!node) { return; }

            // error if the node is not even in this list
            if (this != node.list) { throw new Error("Removing a node from a list it doesn't belong to!"); }

            --count;
            if (first == node)
            {
                first = node.next;
            }
            if (last == node)
            {
                last = node.prev;
            }
            if (node.next) { node.next.prev = node.prev; }
            if (node.prev) { node.prev.next = node.next; }

            ObjectPool.returnLinkedListNode(node);
            if (quickLookUpTable) { delete quickLookUpTable[node.value]; }
        }

        /**
         * Remove the first item int he list that matches the given value
         */
        public function removeByValue(value:*):void
        {
            remove(find(value));
        }

        /**
         * add an item to front of the list
         */
        public function push_front(value:*):void
        {
            ++count;
            var node:LinkedListNode = ObjectPool.getLinkedListNode();
            node.list = this;
            node.value = value;
            node.next = first;
            node.prev = null;
            if (first) { first.prev = node; }
            first = node;
            if (last == null) { last = node; }
            if (quickLookUpTable) { quickLookUpTable[value] = node; }
        }

        /**
         * add an item to back of the list
         */
        public function push_back(value:*):void
        {
            var node:LinkedListNode = ObjectPool.getLinkedListNode();
            node.list = this;
            node.value = value;
            node.next = null;
            node.prev = last;
            if (last) { last.next = node; }
            last = node;
            if (first == null) { first = node; }
            if (quickLookUpTable) { quickLookUpTable[value] = node; }
            ++count;
        }

        /**
         * if you've already got it in LinkedListNode format for some reason, you can add it using this
         * MAKE SURE NO OTHER LIST OWNS IT
         */
        public function push_node_back(node:LinkedListNode):void
        {
            node.list = this;
            node.next = null;
            node.prev = last;
            if (last) { last.next = node; }
            last = node;
            if (first == null) { first = node; }
            if (quickLookUpTable) { quickLookUpTable[node.value] = node; }
            ++count;
        }

        /**
         * if you've already got it in LinkedListNode format for some reason, you can add it using this
         * MAKE SURE NO OTHER LIST OWNS IT
         */
        public function push_node_front(node:LinkedListNode):void
        {
            ++count;
            node.list = this;
            node.next = first;
            node.prev = null;
            if (first) { first.prev = node; }
            first = node;
            if (last == null) { last = node; }
            if (quickLookUpTable) { quickLookUpTable[node.value] = node; }
        }

        /**
         * remove the item at the front of the list and return its value
         */
        public function pop_front():*
        {
            var popped:LinkedListNode = first,
                value:*;

            if (popped)
            {
                if (first == last )
                {
                    first = last = null;
                    count = 0;
                    value = popped.value;
                    ObjectPool.returnLinkedListNode(popped);
                }
                else
                {
                    --count;
                    first = popped.next;
                    if (first) { first.prev = null; }

                    value = popped.value;
                    ObjectPool.returnLinkedListNode(popped);
                    if (quickLookUpTable) { delete quickLookUpTable[popped.value]; }

                    value = popped.value;
                    ObjectPool.returnLinkedListNode(popped);
                }
            }

            return value;
        }

        /**
         * remove the item at the back of the list and return its value
         */
        public function pop_back():*
        {
            var popped:LinkedListNode = last,
                value:*;

            if (popped)
            {
                if (first == last)
                {
                    first = last = null;
                    count = 0;
                    value = popped.value;
                    ObjectPool.returnLinkedListNode(popped);
                }
                else
                {
                    --count;
                    last = popped.prev;
                    if (last) { last.next = null; }
                    if (quickLookUpTable) { delete quickLookUpTable[popped.value]; }
                    value = popped.value;
                    ObjectPool.returnLinkedListNode(popped);
                }
            }

            return value;
        }

        /**
         * empty out the LinkedList
         */
        public function clear():void
        {
            var node:LinkedListNode = first;
            var next:LinkedListNode;
            while (node)
            {
                next = node.next;
                if (quickLookUpTable) { delete quickLookUpTable[node.value]; }
                ObjectPool.returnLinkedListNode(node);
                node = next;
            }
            first = last = null;
            count = 0;
        }
    }

}
