package stego.gameobject
{
    import flash.display.BitmapData;
    import flash.geom.Point;
    import flash.geom.Rectangle;
    import stego.GameInterface;
    import stego.GameObject;
    import stego.util.collision.CollisionInfo;
    import stego.util.collision.CollisionUtil;
    import stego.util.ObjectPool;

    /**
     * GameObject for tile maps
     * Set the five properties, then start calling setTile.
     * @author Mike Meyer
     */
    public class TileMap extends CanvasGameObject
    {
        /** image data from which to get the images for each tile. */
        public function get tilesetImage():BitmapData { return _tilesetImage; }
        public function set tilesetImage(value:BitmapData):void
        {
            _tilesetImage = value;
            _tilesetColumns = _tilesetImage.width / _tileWidth;
            _tilesetRows = _tilesetImage.height / _tileHeight;
            _maxTileIndex = _tilesetColumns * _tilesetRows;
            _needsTilesRedrawnToBuffer = true;
        }
        protected var _tilesetImage:BitmapData;

        /** width of each tile. */
        public function get tileWidth():uint { return _tileWidth; }
        public function set tileWidth(value:uint):void
        {
            _tileWidth = value;
            _tilesetColumns = _tilesetImage.width / _tileWidth;
            _maxTileIndex = _tilesetColumns * _tilesetRows;
            _needsTilesRedrawnToBuffer = true;
        }
        protected var _tileWidth:uint;

        /** height of each tile. */
        public function get tileHeight():uint { return _tileHeight; }
        public function set tileHeight(value:uint):void
        {
            _tileHeight = value;
            _tilesetRows = _tilesetImage.height / _tileHeight;
            _maxTileIndex = _tilesetColumns * _tilesetRows;
            _needsTilesRedrawnToBuffer = true;
        }
        protected var _tileHeight:uint;

        /** width of the entire tileset image */
        override public function set width(value:Number):void
        {
            if (_canvas.width > value)
            {
                _needsTilesRedrawnToBuffer = true;
            }
            super.width = value;
        }

        /** height of the entire tileset image */
        override public function set height(value:Number):void
        {
            if (_canvas.height > value)
            {
                _needsTilesRedrawnToBuffer = true;
            }
            super.height = value;
        }

        /** whether to check for collisions against the edges of the tileset */
        public var checkMapEdgeCollisions:Boolean = false;

        /** 2d array of tile info */
        protected var _tileData:Array;

        /** how many columns are in the tileset source image */
        protected var _tilesetColumns:uint;
        /** how many rows are in the tileset source image */
        protected var _tilesetRows:uint;
        /** highest valid tile index */
        protected var _maxTileIndex:uint;

        /** whether the currently drawn graphics are out of date and need updating */
        protected var _needsTilesRedrawnToBuffer:Boolean;

        /**
         * optimization. copyTileIntoBuffer needs a rect and getting/creating
         * one would be a significant portion of the function's cost.
         * it is called a lot so let's give it a dedicated tempRect
         */
        protected static var _tempRect:Rectangle = new Rectangle();

        /**
         * initial setup and preparation for reuse
         */
        override public function init(game:GameInterface):void
        {
            super.init(game);
            _maxTileIndex = 0;
            _tileData = [];
            _tilesetColumns = 0;
            _tilesetRows = 0;
            _tileHeight = 16;
            _tileWidth = 16;
            _tilesetImage = null;
            _needsTilesRedrawnToBuffer = true;
        }

        /**
         * called after all the final positions have been figured
         * just before drawing
         */
        override public function preDraw(timeDelta:Number):void
        {
            super.preDraw(timeDelta);

            if (_needsTilesRedrawnToBuffer)
            {
                updateTileMap();
                _needsTilesRedrawnToBuffer = false;
            }
        }

        /**
         * clear a tile
         */
        public function clearTile(column:uint, row:uint):void
        {
            var tileData:Object = getTileData(column, row, false);
            if (tileData)
            {
                setTileInternal(tileData, -1);
            }
        }

        /**
         * set which index to use for a tile. -1 for empty.
         */
        public function setTile(column:uint, row:uint, index:int):void
        {
            setTileInternal(getTileData(column, row, true), index);
        }

        /**
         * Helper for setTile. actually performs the work after setTile finds/creates the tile
         * @param	tileData
         * @param	index
         */
        protected function setTileInternal(tileData:Object, index:int):void
        {
            tileData.index = index;

            if (_tilesetImage)
            {
                copyTileIntoBuffer(tileData);
            }
        }

        /**
         * get the data object to go with the tile at the given location.
         * @param   createIfNeeded  If true, this will create the tile if it doesn't exist yet
         */
        protected function getTileData(column:uint, row:uint, createIfNeeded:Boolean):Object
        {
            var rowData:Array = _tileData[row];
            if (!rowData)
            {
                if (!createIfNeeded) { return null; }
                rowData = _tileData[row] = [];
            }
            var tileData:Object = rowData[column];
            if (!tileData)
            {
                if (!createIfNeeded) { return null; }
                tileData = rowData[column] = { index: -1, row:row, col:column };
            }
            return tileData;
        }

        /**
         * Copy a tile image into a spot on the buffer
         * does not check that it is safe to do so
         * @param	tileData    tileData for the tile we are rendering
         */
        protected function copyTileIntoBuffer(tileData:Object):void
        {
            var tempPoint:Point = ObjectPool.tempPoint;
            var index:int = tileData.index;

            if (index >= 0)
            {
                var column:Number = tileData.col;
                var row:Number = tileData.row;
                _tempRect.x = (index % _tilesetColumns) * _tileWidth;
                _tempRect.y = uint(index / _tilesetColumns) * _tileHeight;
                _tempRect.width = _tileWidth;
                _tempRect.height = _tileHeight;
                tempPoint.x = column * _tileWidth;
                tempPoint.y = row * _tileHeight;
                internalCopyTileIntoBuffer(tileData, _tempRect, tempPoint);
            }
            else
            {
                _tempRect.x = column * _tileWidth;
                _tempRect.y = row * _tileHeight;
                _tempRect.width = _tileWidth;
                _tempRect.height = _tileHeight;
                internalClearTileInBuffer(tileData, _tempRect);
            }
        }

        /**
         * helper for copyTileIntoBuffer.  extracted for easier overriding
         * @param	tileData        tileData for the tile we are rendering
         * @param	sourceRect      source area of the tile image
         * @param   destPoint       where in the buffer to render it at
         */
        protected function internalCopyTileIntoBuffer( tileData:Object,
                                                      sourceRect:Rectangle,
                                                      destPoint:Point ):void
        {
            _canvas.copyPixels( _tilesetImage,
                                sourceRect,
                                destPoint );
        }

        /**
         * helper for copyTileIntoBuffer.  extracted for easier overriding
         * @param	tileData    tileData for the tile we are rendering
         * @param	rect        area in the buffer for that tile
         */
        protected function internalClearTileInBuffer( tileData:Object,
                                                      rect:Rectangle ):void
        {
            _canvas.fillRect(rect,0);
        }

        /**
         * update the big image used for the tile map graphics
         */
        protected function updateTileMap():void
        {
            for (var row:uint = 0; row < _tileData.length; ++row)
            {
                var rowData:Array = _tileData[row];
                if (!rowData) { continue; }
                for (var column:uint = 0; column < rowData.length; ++column)
                {
                    var tileData:Object = rowData[column];
                    if (!tileData ||  tileData.index<0) { continue; }
                    copyTileIntoBuffer(tileData);
                }
            }
            _needsTilesRedrawnToBuffer = false;
        }

        /**
         * clean up
         */
        override public function shutdown():void
        {
            super.shutdown();
            _tilesetImage = null;
            _tileData = null;
        }

        /**
         * detect and react to collisions with a regular GameObject
         * @param   go          object to collide against the tilemap
         * @param   callback    for each tile that go collides with, this will
         *                      get called and sent the tileData object (that
         *                      you can modify), tileMap, game object, and
         *                      collisionIinfo.  so it should look like
         *                      function tileCollisionCallback(tileData:Object, tileMap:TileMap, go:GameObject, ci:CollisionInfo):void
         */
        public function collide(go:GameObject, callback:Function = null):void
        {
            var bounds:Rectangle = go.getCollisionBounds( ObjectPool.getObject(Rectangle) ),
                offsetLeft:Number = (bounds.x - x), // distance from the left edge of the tilemap to the left edge of the object
                offsetTop:Number = (bounds.y - y), // distance from the top of the tilemap to the top of the object
                offsetRight:Number = offsetLeft + bounds.width, // distance from the left edge of the tilemap to the right edge of the object
                offsetBottom:Number = offsetTop + bounds.height; // distance from the top of the tilemap to the bottom of the object

            if ( (offsetLeft   < width)  &&
                 (offsetTop    < height) &&
                 (offsetRight  > 0.0)    &&
                 (offsetBottom > 0.0) )
            {
                var firstRow:uint = (offsetTop < 0.0) ? 0 : uint(offsetTop / _tileHeight);
                var lastRow:uint = (offsetBottom < 0.0) ? 0 : uint(offsetBottom / _tileHeight);
                lastRow = (offsetBottom == lastRow) ? lastRow : lastRow + 1;

                var firstCol:uint = (offsetLeft < 0.0) ? 0 : uint(offsetLeft / _tileWidth);
                var lastCol:uint = (offsetRight < 0.0) ? 0 : uint(offsetRight / _tileWidth);
                lastCol = (offsetRight == lastCol) ? lastCol : lastCol + 1;

                var tileRect:Rectangle = ObjectPool.getObject(Rectangle),
                    row:uint,
                    col:uint,
                    row_i:uint,
                    row_i_max:uint = lastRow-firstRow,
                    col_i:uint,
                    col_i_max:uint = lastCol-firstCol,
                    rowData:Array,
                    tileData:Object,
                    ci:CollisionInfo,
                    // look at its motion and first check the tiles that it would hit earliest;
                    goRight:Boolean = (go.dx >= 0),
                    goDown:Boolean = (go.dy >= 0);

                for (row_i=0; row_i<=row_i_max; ++row_i)
                {
                    row = goDown ? firstRow + row_i : lastRow - row_i;
                    rowData = _tileData[row];
                    if (!rowData) { continue; }
                    for (col_i=0; col_i <= col_i_max; ++col_i)
                    {
                        col = goRight ? firstCol+col_i : lastCol-col_i;
                        tileData = rowData[col];
                        if ( !tileData || (tileData.index<0) ) { continue; }

                        tileRect.x = (col * _tileWidth) + x;
                        tileRect.y = (row * _tileHeight) + y;
                        tileRect.width = _tileWidth;
                        tileRect.height = _tileHeight;

                        ci = checkTileCollision(tileData, tileRect, bounds);
                        if (ci)
                        {
                            // let the character know he hit something
                            go.collisionOccurred(ci, this);
                            ci.normal.x *= -1;
                            ci.normal.y *= -1;
                            this.collisionOccurred(ci, go);
                            ci.normal.x *= -1;
                            ci.normal.y *= -1;

                            // update values and continue on checking the next tile
                            go.getCollisionBounds( bounds );
                            offsetLeft = (bounds.x - x); // distance from the left edge of the tilemap to the left edge of the object
                            offsetTop = (bounds.y - y); // distance from the top of the tilemap to the top of the object
                            offsetRight = offsetLeft + bounds.width; // distance from the left edge of the tilemap to the right edge of the object
                            offsetBottom = offsetTop + bounds.height; // distance from the top of the tilemap to the bottom of the object

                            // call the callack, if we were given one
                            if (null != callback)
                            {
                                callback(tileData, this, go, ci);
                            }
                        }
                    }
                }

                // check map edges (optionally)
                if (checkMapEdgeCollisions)
                {
                    ci = ObjectPool.getObject(CollisionInfo);
                    if (offsetLeft < 0)
                    {
                        ci.t = 0;
                        ci.normal.x = 1;
                        ci.normal.y = 0;
                        ci.location.x = x;
                        ci.location.y = 0.5 * (bounds.top + bounds.bottom);

                        // let the character know he hit something
                        go.collisionOccurred(ci, this);
                        ci.normal.x *= -1;
                        ci.normal.y *= -1;
                        this.collisionOccurred(ci, go);

                        // update values and continue checking
                        bounds = go.getCollisionBounds( bounds );
                        offsetLeft = (bounds.x - x); // distance from the left edge of the tilemap to the left edge of the object
                        offsetTop = (bounds.y - y); // distance from the top of the tilemap to the top of the object
                        offsetRight = offsetLeft + bounds.width; // distance from the left edge of the tilemap to the right edge of the object
                        offsetBottom = offsetTop + bounds.height; // distance from the top of the tilemap to the bottom of the object
                    }
                    if (offsetRight > width)
                    {
                        ci.t = 0;
                        ci.normal.x = -1;
                        ci.normal.y = 0;
                        ci.location.x = width;
                        ci.location.y = 0.5 * (bounds.top + bounds.bottom);

                        // let the character know he hit something
                        go.collisionOccurred(ci, this);
                        ci.normal.x *= -1;
                        ci.normal.y *= -1;
                        this.collisionOccurred(ci, go);

                        // update values and continue checking
                        bounds = go.getCollisionBounds( bounds );
                        offsetLeft = (bounds.x - x); // distance from the left edge of the tilemap to the left edge of the object
                        offsetTop = (bounds.y - y); // distance from the top of the tilemap to the top of the object
                        offsetRight = offsetLeft + bounds.width; // distance from the left edge of the tilemap to the right edge of the object
                        offsetBottom = offsetTop + bounds.height; // distance from the top of the tilemap to the bottom of the object
                    }
                    if (offsetTop < 0)
                    {
                        ci.t = 0;
                        ci.normal.x = 0;
                        ci.normal.y = 1;
                        ci.location.x = 0.5 * (bounds.left + bounds.right);
                        ci.location.y = y;

                        // let the character know he hit something
                        go.collisionOccurred(ci, this);
                        ci.normal.x *= -1;
                        ci.normal.y *= -1;
                        this.collisionOccurred(ci, go);

                        // update values and continue checking
                        bounds = go.getCollisionBounds( bounds );
                        offsetLeft = (bounds.x - x); // distance from the left edge of the tilemap to the left edge of the object
                        offsetTop = (bounds.y - y); // distance from the top of the tilemap to the top of the object
                        offsetRight = offsetLeft + bounds.width; // distance from the left edge of the tilemap to the right edge of the object
                        offsetBottom = offsetTop + bounds.height; // distance from the top of the tilemap to the bottom of the object
                    }
                    if (offsetBottom > height)
                    {
                        ci.t = 0;
                        ci.normal.x = 0;
                        ci.normal.y = -1;
                        ci.location.x = 0.5 * (bounds.left + bounds.right);
                        ci.location.y = height;

                        // let the character know he hit something
                        go.collisionOccurred(ci, this);
                        ci.normal.x *= -1;
                        ci.normal.y *= -1;
                        this.collisionOccurred(ci, go);

                        // update values and continue checking
                        bounds = go.getCollisionBounds( bounds );
                        offsetLeft = (bounds.x - x); // distance from the left edge of the tilemap to the left edge of the object
                        offsetTop = (bounds.y - y); // distance from the top of the tilemap to the top of the object
                        offsetRight = offsetLeft + bounds.width; // distance from the left edge of the tilemap to the right edge of the object
                        offsetBottom = offsetTop + bounds.height; // distance from the top of the tilemap to the bottom of the object
                    }
                    ObjectPool.returnObject(CollisionInfo, ci);
                }
                ObjectPool.returnObject(Rectangle, tileRect);
            }
            ObjectPool.returnObject(Rectangle, bounds);
        }

        /**
         * helper for collide.  extracted for easier overriding.
         * @param   tileData    tileData for the tile we are checking
         * @param   tileRect    collision bounds for the tile
         * @param   objectRect  collision bounds for the object
         */
        protected function checkTileCollision(tileData:Object, tileRect:Rectangle, objectRect:Rectangle):CollisionInfo
        {
            return CollisionUtil.checkRectCollision(objectRect, tileRect);
        }

        /**
         * tilemaps never move in response to collisions
         */
        override public function collisionOccurred(ci:CollisionInfo, object:GameObject):void
        {
        }
    }
}
