﻿package stego.util
{
    import flash.geom.Point;
    import flash.geom.Rectangle;
    import stego.collections.LinkedList;
    import stego.collections.LinkedListNode;
    import stego.GameInterface;
    import stego.GameObject;
    import stego.debug.Profiler;
    import stego.util.collision.CollisionInfo;
    import stego.util.collision.CollisionUtil;
    import stego.util.scene.GameObjectHolder;
    import stego.util.scene.Scene;
    import stego.util.scene.GameObjectHolder;

    /**
     * Class to handle the physics simulation and collision checks.
     * to use it, add layers and collisionPairs to tell it what layers
     * get processed and what ones collide with each other.
     * @author Mike Meyer
     */
    public class PhysicsUtil
    {
        /**
         * list of layers whose objects will get physics'd
         */
        private var _layers:/*GameObjectHolder*/LinkedList;

        /**
         * list of pairs whose objects will get physics'd
         */
        private var _collisionPairs:/*Pair*/LinkedList;

        /**
         * get/set the acceleration due to gravity in this game
         */
        public var gravity:Number = 2000;

        /**
         * get/set a maximum fall speed for this game
         */
        public var maxFallSpeed:Number = 600;

        /**
         * game instance that this belongs to
         */
        private var _game:GameInterface;

        /**
         * set up the physics util
         */
        public function init(game:GameInterface):void
        {
            _game = game;
            _layers = new LinkedList(true);
            _collisionPairs = new LinkedList();
        }
        /**
         * clean up the physics util
         */
        public function shutdown():void
        {
            _collisionPairs.clear();
            _layers.clear();
            _game = null;
        }

        /**
         *
         */
        public function update(timeDelta:Number):void
        {
            var layerNode:LinkedListNode;
            var layer:GameObjectHolder;
            var objectNode:LinkedListNode;
            var object:GameObject;

            // apply gravity to objects in physics'd layers
            for (layerNode = _layers.first; layerNode != null; layerNode = layerNode.next)
            {
                layer = layerNode.value;
                for (objectNode = layer.objects.first; objectNode != null; objectNode = objectNode.next)
                {
                    object = objectNode.value;
                    if (!object.physicsEnabled) { continue; }
                    applyGravityTo(object);
                }
            }
        }

        /**
         *
         */
        public function integrate(timeDelta:Number):void
        {
        }

        /**
         *
         */
        public function preCollision(timeDelta:Number):void
        {
        }

        /**
         *
         */
        public function collision(timeDelta:Number):void
        {
            var layer1Node:LinkedListNode;
            var layer1:GameObjectHolder;
            var object1Node:LinkedListNode;
            var object1:GameObject;
            var layer2:GameObjectHolder;
            var object2Node:LinkedListNode;
            var object2:GameObject;

            // check pairs for collisions
            for (var pairNode:LinkedListNode = _collisionPairs.first; pairNode != null; pairNode = pairNode.next)
            {
                var pair:Pair = pairNode.value;
                layer1 = pair.a;
                layer2 = pair.b;
                for (object1Node = layer1.objects.first; object1Node!=null; object1Node = object1Node.next)
                {
                    object1 = object1Node.value;
                    if (!object1.physicsEnabled) { continue; }

                    // don't double-compare if it is the same layer
                    object2Node = (layer1==layer2) ? object1Node.next : layer2.objects.first;

                    for (; object2Node != null; object2Node = object2Node.next)
                    {
                        object2 = object2Node.value;
                        if (!object2.physicsEnabled) { continue; }
                        checkCollision(object1, object2);
                    }
                }
            }
        }

        /**
         *
         */
        public function preDraw(timeDelta:Number):void
        {
        }

        /**
         * detect and react to the collisions
         */
        protected function checkCollision(go1:GameObject, go2:GameObject):void
        {
            var ci:CollisionInfo = CollisionUtil.checkAabbCollision(go1, go2, _game.time.timeDelta);

            if (null != ci)
            {
                // let the objects know they hit each other
                go1.collisionOccurred(ci, go2);
                ci.normal.x *= -1;
                ci.normal.y *= -1;
                go2.collisionOccurred(ci, go1);
            }
        }

        /**
         * apply gravity to the object
         */
        protected function applyGravityTo(go:GameObject):void
        {
            if (go.affectedByGravity)
            {
                go.dy += gravity * _game.time.timeDelta;

                if (go.dy > maxFallSpeed)
                {
                    go.dy = maxFallSpeed;
                }
            }
        }

        /**
         * add a layer to the physics simulation
         * @param	layer   layer to add
         */
        public function addLayer(layer:GameObjectHolder):void
        {
            _layers.push_back(layer);
        }

        /**
         * remove a layer from the physics simulation
         * @param	layer   layer to remove
         */
        public function removeLayer(layer:GameObjectHolder):void
        {
            _layers.removeByValue(layer);
        }

        /**
         * remove all layers from the physics simulation
         */
        public function removeAllLayers():void
        {
            _layers.clear();
        }

        /**
         * tell the physicsUtil to check collisions between objects in a and
         * objects in b.  (they can be the same layer)
         * Note that the collision code prefers to move A
         */
        public function addCollisionPair(a:GameObjectHolder, b:GameObjectHolder):void
        {
            _collisionPairs.push_back(new Pair(a, b));
        }

        /**
         * remove all registered collision pairs
         */
        public function removeAllCollisionPairs():void
        {
            _collisionPairs.clear();
        }
    }

}
import stego.util.scene.GameObjectHolder;
import stego.util.scene.GameObjectHolder;

/**
 * for associating pairs of layers to do collision checks on
 */
class Pair
{
    public function Pair(a:GameObjectHolder = null, b:GameObjectHolder = null) { this.a = a; this.b = b; }
    public var a:GameObjectHolder;
    public var b:GameObjectHolder;
}
