﻿package stego.util.collision
{
    import flash.display.DisplayObject;
    import flash.geom.Point;
    import flash.geom.Rectangle;
    import stego.GameInterface;
    import stego.GameObject;
    import stego.debug.Profiler;
    import stego.util.collision.CollisionInfo;
    import stego.util.MathUtil;
    import stego.util.ObjectPool;
    import stego.ui.TextField;

    /**
     * handy functions for doing collision checks
     * @author Mike Meyer
     */
    public class CollisionUtil
    {
        // the farthest (relative) distance we want to go between collision tests
        private static const COLLISION_STEP_SIZE_MAX:Number = 5;

        /**
         * check for collisions between two moving GameObjects
         * returns null if they did not collide
         */
        public static function checkAabbCollision(objectA:GameObject, objectB:GameObject, elapsedTime:Number):CollisionInfo
        {
            var objectABounds:Rectangle = objectA.getCollisionBounds();
            var objectBBounds:Rectangle = objectB.getCollisionBounds();

            var aMotionRelativeToB_x:Number = (objectA.dx - objectB.dx) * elapsedTime;
            var aMotionRelativeToB_y:Number = (objectA.dy - objectB.dy) * elapsedTime;

            if (!quickCollisionPossibilityCheck( objectABounds,
                                                 objectBBounds,
                                                 aMotionRelativeToB_x,
                                                 aMotionRelativeToB_y ) )
            {
                return null;
            }

            return CollisionUtil.internalCheckAabbCollision(
                objectABounds,
                objectBBounds,
                aMotionRelativeToB_x,
                aMotionRelativeToB_y );
        }

        /**
         * check for collisions between two rects
         * @param   a       rect for one of the objects
         * @param   b       rect for the other
         */
        public static function checkRectCollision( a:Rectangle,
                                                   b:Rectangle ):CollisionInfo
        {
            // quick out if they are not overlapping
            if (!a.intersects(b))
            {
                return null;
            }

            var aCenterX:Number = a.left + (0.5 * a.width),
                aCenterY:Number = a.top + (0.5 * a.height),
                bCenterX:Number = b.left + (0.5 * b.width),
                bCenterY:Number = b.top + (0.5 * b.height),
                xOverlap:Number,
                yOverlap:Number,
                absXOverlap:Number,
                absYOverlap:Number;

            if (aCenterX > bCenterX)
            {
                xOverlap = a.left - b.right;
            }
            else if (aCenterX < bCenterX)
            {
                xOverlap = b.left - a.right;
            }
            else
            {
                xOverlap = 0; // the centers line up.  don't even check x collisions
            }

            if (aCenterY > bCenterY)
            {
                yOverlap = a.top - b.bottom;
            }
            else if (aCenterY < bCenterY)
            {
                yOverlap = b.top - a.bottom;
            }
            else
            {
                yOverlap = 0; // the centers line up.  don't even check y collisions
            }

            absXOverlap = (xOverlap < 0.0) ? -xOverlap : xOverlap;
            absYOverlap = (yOverlap < 0.0) ? -yOverlap : yOverlap;

            var ci:CollisionInfo = null;
            if ( (yOverlap != 0) &&
                 ((absYOverlap < absXOverlap) || (absXOverlap==0.0)))
            {
                ci = new CollisionInfo();
                ci.location.x = aCenterX;
                ci.location.y = (aCenterY > bCenterY) ? b.bottom : b.top;
                ci.normal.x = 0;
                ci.normal.y = (aCenterY > bCenterY) ? 1.0 : -1.0;
            }
            else if (xOverlap != 0)
            {
                ci = new CollisionInfo();
                ci.location.x = (aCenterX > bCenterX) ? b.right : b.left;
                ci.location.y = aCenterY;
                ci.normal.x = (aCenterX > bCenterX) ? 1.0 : -1.0;
                ci.normal.y = 0;
            }

            return ci;
        }

        /**
         * check for collisions between two potentially moving rects
         */
        private static function internalCheckAabbCollision(
            objectABounds:Rectangle,
            objectBBounds:Rectangle,
            aMotionRelativeToB_x:Number,
            aMotionRelativeToB_y:Number ):CollisionInfo
        {
            var ci:CollisionInfo = null;
            var tempRect:Rectangle = ObjectPool.getObject(Rectangle);

            var dist:Number = Math.sqrt( (aMotionRelativeToB_x*aMotionRelativeToB_x) + (aMotionRelativeToB_y*aMotionRelativeToB_y) );
            var steps:Number = 1 + Math.floor(dist / COLLISION_STEP_SIZE_MAX);

            for (var step:Number = 0; step <= steps; ++step)
            {
                // keep the very edges of things from causing unwanted collisions
                const OVERLAP_FUDGE:Number = 0.5;

                var scale:Number = (step) / steps;
                tempRect.x = objectABounds.x + OVERLAP_FUDGE;
                tempRect.y = objectABounds.y + OVERLAP_FUDGE;
                tempRect.width = objectABounds.width - 2*OVERLAP_FUDGE;
                tempRect.height = objectABounds.height - 2*OVERLAP_FUDGE;
                tempRect.offset(aMotionRelativeToB_x * scale, aMotionRelativeToB_y * scale);

                if (tempRect.intersects(objectBBounds))
                {
                    ci = createCollisionInfoForAabbCollision( tempRect,
                                                              objectBBounds,
                                                              aMotionRelativeToB_x,
                                                              aMotionRelativeToB_y );
                    ci.t = step / steps;

                    if (ci.normal.length == 0)
                    {
                        throw new Error("Some kind of error in createCollisionInfoForAabbCollision");
                        ci = null;
                    }

                    break;
                }
            }

            ObjectPool.returnObject(Rectangle, tempRect);
            return ci;
        }

        /**
         * quick check for whether there is any chance of collision
         */
        private static function quickCollisionPossibilityCheck(
            objectABounds:Rectangle,
            objectBBounds:Rectangle,
            aMotionRelativeToB_x:Number,
            aMotionRelativeToB_y:Number ):Boolean
        {
            // too far left
            var aMaxX:Number = Math.max(objectABounds.right, objectABounds.right + aMotionRelativeToB_x);
            if (aMaxX < objectBBounds.left) { return false; }

            // too far right
            var aMinX:Number = Math.min(objectABounds.left, objectABounds.left + aMotionRelativeToB_x);
            if (aMinX > objectBBounds.right) { return false; }

            // too far up
            var aMaxY:Number = Math.max(objectABounds.bottom, objectABounds.bottom + aMotionRelativeToB_y);
            if (aMaxY < objectBBounds.top) { return false; }

            // too far down
            var aMinY:Number = Math.min(objectABounds.top, objectABounds.top + aMotionRelativeToB_y);
            if (aMinY > objectBBounds.bottom) { return false; }

            return true;
        }

        /**
         * used internally by CheckAabbCollision
         */
        private static function createCollisionInfoForAabbCollision(
            objectA:Rectangle,
            objectB:Rectangle,
            xmove:Number, ymove:Number):CollisionInfo
        {
            var a_center:Point = MathUtil.getRectCenter(objectA);
            var b_center:Point = MathUtil.getRectCenter(objectB);
            var ci:CollisionInfo = new CollisionInfo();
            // defaults
            ci.location.x = a_center.x;
            ci.location.y = a_center.y;
            ci.normal.x = 0.0;
            ci.normal.y = 0.0;

            var xdiff:Number = (xmove > 0.0) ? (objectA.right - objectB.left) : (objectB.right - objectA.left);
            var ydiff:Number = (ymove > 0.0) ? (objectA.bottom - objectB.top) : (objectB.bottom - objectA.top);

            var x_first:Boolean = ( Math.abs(xmove) > Math.abs(ymove) );
            var checkNextAxis:Boolean = true;

            if (x_first && checkNextAxis)
            {
                if ( xmove > 0 )
                {
                    if ( objectB.containsPoint( new Point( objectA.right, a_center.y ) ) )
                    {
                        ci.location.x = objectB.left;
                        ci.normal.x = -1;
                        ci.normal.y =  0;
                        checkNextAxis = false;
                    }
                }
                else
                {
                    if ( objectB.containsPoint( new Point( objectA.left, a_center.y ) ) )
                    {
                        ci.location.x = objectB.right;
                        ci.normal.x =  1;
                        ci.normal.y =  0;
                        checkNextAxis = false;
                    }
                }
            }

            if (checkNextAxis)
            {
                if (ymove > 0)
                {
                    if ( objectB.containsPoint( new Point( a_center.x, objectA.bottom ) ) )
                    {
                        ci.location.y = objectB.top;
                        ci.normal.x =  0;
                        ci.normal.y = -1;
                        checkNextAxis = false;
                    }
                }
                else
                {
                    if ( objectB.containsPoint( new Point( a_center.x, objectA.top ) ) )
                    {
                        ci.location.y = objectB.bottom;
                        ci.normal.x =  0;
                        ci.normal.y =  1;
                        checkNextAxis = false;
                    }
                }
            }

            if (!x_first && checkNextAxis)
            {
                if ( xmove > 0 )
                {
                    if ( objectB.containsPoint( new Point( objectA.right, a_center.y ) ) )
                    {
                        ci.location.x = objectB.left;
                        ci.normal.x = -1;
                        ci.normal.y =  0;
                        checkNextAxis = false;
                    }
                }
                else
                {
                    if ( objectB.containsPoint( new Point( objectA.left, a_center.y ) ) )
                    {
                        ci.location.x = objectB.right;
                        ci.normal.x =  1;
                        ci.normal.y =  0;
                        checkNextAxis = false;
                    }
                }
            }

            // do yet more checking now that the "obvious" cases are handled.
            if (checkNextAxis)
            {
                if (x_first && checkNextAxis)
                {
                    if (xmove > 0)
                    {
                        if ( (objectA.right > objectB.left) && (objectA.right < objectB.right) )
                        {
                            ci.location.x = objectB.left;
                            ci.normal.x = -1;
                            ci.normal.y =  0;
                            checkNextAxis = false;
                        }
                    }
                    else
                    {
                        if ( (objectA.left > objectB.left) && (objectB.left < objectB.right)
                          && ( (objectA.left - xmove) >= (objectB.right) ) ) // make sure it was this frame that it crossed the edge
                        {
                            ci.location.x = objectB.right;
                            ci.normal.x =  1;
                            ci.normal.y =  0;
                            checkNextAxis = false;
                        }
                    }
                }

                if (checkNextAxis)
                {
                    if (ymove > 0)
                    {
                        if ( (objectA.bottom > objectB.top) && (objectA.bottom < objectB.bottom) )
                        {
                            ci.location.y = objectB.top;
                            ci.normal.x =  0;
                            ci.normal.y = -1;
                            checkNextAxis = false;
                        }
                    }
                    else
                    {
                        if ( (objectA.top > objectB.top) && (objectB.top < objectB.bottom) )
                        {
                            ci.location.y = objectB.bottom;
                            ci.normal.x =  0;
                            ci.normal.y =  1;
                            checkNextAxis = false;
                        }
                    }
                }

                if (!x_first && checkNextAxis)
                {
                    if (xmove > 0)
                    {
                        if ( (objectA.right > objectB.left) && (objectA.right < objectB.right) )
                        {
                            ci.location.x = objectB.left;
                            ci.normal.x = -1;
                            ci.normal.y =  0;
                            checkNextAxis = false;
                        }
                    }
                    else
                    {
                        if ( (objectA.left > objectB.left) && (objectB.left < objectB.right) )
                        {
                            ci.location.x = objectB.right;
                            ci.normal.x =  1;
                            ci.normal.y =  0;
                            checkNextAxis = false;
                        }
                    }
                }
            }

            // something really bad happened above and it was unclear from the
            // way they overlap which way to go. just use how he was moving to
            // do a best guess.
            if (checkNextAxis)
            {
                if (x_first)
                {
                    if (xmove > 0)
                    {
                        ci.location.x = objectB.left;
                        ci.normal.x = -1;
                        ci.normal.y =  0;
                        checkNextAxis = false;
                    }
                    else
                    {
                        ci.location.x = objectB.right;
                        ci.normal.x =  1;
                        ci.normal.y =  0;
                        checkNextAxis = false;
                    }
                }
                else
                {
                    if (ymove > 0)
                    {
                        ci.location.y = objectB.top;
                        ci.normal.x =  0;
                        ci.normal.y = -1;
                        checkNextAxis = false;
                    }
                    else
                    {
                        ci.location.y = objectB.bottom;
                        ci.normal.x =  0;
                        ci.normal.y =  1;
                        checkNextAxis = false;
                    }
                }
            }

            ci.t = 0.0; // needs to be set outside this function

            return ci;
        }
    }
}
