package stego.util.graphics
{
    import flash.display.Bitmap;
    import flash.display.BitmapData;
    import flash.geom.ColorTransform;
    import flash.geom.Matrix;
    import flash.geom.Point;
    import flash.geom.Rectangle;
    import stego.GameInterface;
    import stego.util.ObjectPool;

    /**
     * wraps a Bitmap to more easily draw on it
     * @author Mike Meyer
     */
    public class Canvas
    {
        /** width of the entire tileset image */
        public function get width():int { return _width; }
        public function set width(value:int):void
        {
            if (_width > value)
            {
                _bufferUpToDate = false;
            }
            _width = value;
        }
        protected var _width:int;

        /** height of the entire tileset image */
        public function get height():int { return _height; }
        public function set height(value:int):void
        {
            if (_height > value)
            {
                _bufferUpToDate = false;
            }
            _height = value;
        }
        protected var _height:int;

        /** Bitmap object it draws on */
        public function get target():Bitmap { return _target; }
        public function set target(value:Bitmap):void
        {
            if (_target != value)
            {
                _target = value;
                _bufferUpToDate = false;
            }
        }
        protected var _target:Bitmap;

        /** whether the current buffer needs rebuilding before we can redraw individual tiles */
        protected var _bufferUpToDate:Boolean = false;

        /** whether the canvas has already been cleared */
        protected var _isClear:Boolean = true;

        /**
         * prepare the canvas for use
         */
        public function Canvas(target:Bitmap = null, width:int = 100, height:int = 100):void
        {
            _width = width;
            _height = height;
            _target = target;
        }

        /** no point in new'ing these all the time */
        protected static var _tempPoint:Point = new Point();
        /** no point in new'ing these all the time */
        protected static var _tempMatrix:Matrix = new Matrix();

        /**
         * clear whatever has been drawn to the canvas
         */
        public function clear():void
        {
            if (!_isClear && _target && _target.bitmapData)
            {
                _target.bitmapData.fillRect(_target.bitmapData.rect, 0);
                _isClear = true;
            }
        }

        /**
         * fades out the bitmap contents
         * @param   amount  how much to fade the canvas by. [0-1] 0 is not at all. 1 makes it completely transparent
         */
        public function fade(amount:Number):void
        {
            if (!_isClear && _target && _target.bitmapData)
            {
                var ct:ColorTransform = ObjectPool.getObject(ColorTransform);
                ct.alphaMultiplier = 1 - amount;
                ct.alphaOffset = 0;
                ct.redMultiplier = 1;
                ct.redOffset = 0;
                ct.greenMultiplier = 1;
                ct.greenOffset = 0;
                ct.blueMultiplier = 1;
                ct.blueOffset = 0;
                _target.bitmapData.colorTransform(_target.bitmapData.rect, ct);
                ObjectPool.returnObject(ColorTransform, ct);
            }
        }

        /**
         * Fills the rectangle area of the canvas with the texture.
         * @param	bitmapData      bitmapdata to draw.
         * @param	sourceRect      what part of the source bitmap to use
         * @param	destRect        what part of the canvas to draw to
         * @param	clobberAlpha    If set to true, then instead of doing alpha
         *                          blending, the alpha in the source image
         *                          will be copied into the canvas.
         */
        public function copyPixels(bitmapData:BitmapData, sourceRect:Rectangle, destPoint:Point, clobberAlpha:Boolean = false):void
        {
            resizeBufferIfNeeded();
            _target.bitmapData.copyPixels(bitmapData, sourceRect, destPoint, null, null, !clobberAlpha);
            _isClear = false;
        }
        /**
         * Fills the rectangle area of the canvas with the texture.
         * @param	bitmapData      bitmapdata to draw.
         * @param	sourceRect      what part of the source bitmap to use
         * @param	matrix          matrix to use to transform the image (or null)
         * @param	colorTransform  colorTransform to apply (or null)
         * @param	blendMode       value from the flash.display.BlendMode class for how to blend the image in (or null)
         * @param   smoothing       whether to smooth the image when transformed.
         */
        public function drawBitmap( bitmapData:BitmapData,
                                    sourceRect:Rectangle = null,
                                    matrix:Matrix = null,
                                    colorTransform:ColorTransform = null,
                                    blendMode:String = null,
                                    smoothing:Boolean = false ):void
        {
            resizeBufferIfNeeded();

            if (!sourceRect) { sourceRect = bitmapData.rect; }

            var buffer:BitmapData = ObjectPool.getBitmapData(sourceRect.width, sourceRect.height);
            _tempPoint.x = 0;
            _tempPoint.y = 0;
            buffer.copyPixels(bitmapData, sourceRect, _tempPoint);
            _target.bitmapData.draw(buffer, matrix, colorTransform, blendMode, null, smoothing);
            ObjectPool.returnBitmapData(buffer);

            _isClear = false;
        }

        /**
         * Fills the rectangle area of the canvas with a solid color.
         * @param	rect    area to fill
         * @param   color   ARGB value to fill with
         */
        public function fillRect(rect:Rectangle, color:uint):void
        {
            resizeBufferIfNeeded();
            _target.bitmapData.fillRect(rect, color);
            _isClear = false;
        }

        /**
         * Set just one pixel of the canvas
         * @param	x       x position
         * @param   y       y position
         * @param   color   ARGB color to set it to
         */
        public function setPixel(x:int, y:int, color:uint):void
        {
            resizeBufferIfNeeded();
            _target.bitmapData.setPixel32(x,y,color);
            _isClear = false;
        }

        /**
         * make sure that our buffer is ready to be shown/drawn on
         */
        protected function resizeBufferIfNeeded():void
        {
            if (!_target.bitmapData || !_bufferUpToDate)
            {
                var oldBitmapData:BitmapData = _target.bitmapData;
                _target.bitmapData = new BitmapData(_width, _height, true, 0);
                _tempPoint.x = 0;
                _tempPoint.y = 0;
                if (oldBitmapData)
                {
                    _target.bitmapData.copyPixels(oldBitmapData, oldBitmapData.rect, _tempPoint);
                }
                _bufferUpToDate = true;
            }
        }
    }

}
