﻿package stego.util.music
{
    import flash.errors.IOError;
    import flash.events.Event;
    import flash.events.EventDispatcher;
    import flash.events.IOErrorEvent;
    import flash.media.Sound;
    import flash.media.SoundChannel;
    import flash.media.SoundLoaderContext;
    import flash.media.SoundTransform;
    import flash.net.URLRequest;
    import flash.utils.Timer;
    import stego.util.ArrayUtil;
    import stego.util.music.MusicEvent;
    import stego.util.music.Song;

    /**
     * Class to handle randomizing, looping, and streaming a playlist
     *
     * Example use:
     *
     *      var music:MusicUtil - new Music;
     *      music.Init( [ new Song( "http://8bitcollective.com/items/music/joule-girl_scrotum.mp3",
     *                              "Girl Scrotum",
     *                              "joule",
     *                              "http://8bitcollective.com/music/joule/girl+scrotum+%28OHC%29/"),
     *                    new Song( "http://8bitcollective.com/items/music/Untill%20I%20Sleep.mp3",
     *                              "Until I Sleep",
     *                              "henryhomesweet",
     *                              "http://8bitcollective.com/music/henryhomesweet/Until+I+Sleep/") ] );
     *      music.StartNewMusic();
     *
     * @author Mike Meyer
     */
    public class MusicUtil extends EventDispatcher
    {
        private var music:Sound;
        private var musicChannel:SoundChannel;
        private var pausePosition:Number;

        /**
         * get/set the volume that things play back at
         */
        public function get volume():Number { return _volume; }
        public function set volume(value:Number):void
        {
            if (_volume == value) { return; }

            _volume = value;
            if (musicChannel)
            {
                // get to use a temp var and reassign to have the new settings take effect.  Flash is weird like that.
                var st:SoundTransform = musicChannel.soundTransform;
                st.volume = _volume * _songVolume;
                musicChannel.soundTransform = st;
            }
        }
        private var _volume:Number = 0.5;
        private var _songVolume:Number = 0.5;

        /**
         * currently playing song.  may be null
         */
        public function get currentSong():Song { return _currentSong; }
        private var _currentSong:Song;

        public function get isPlaying():Boolean { return (null!=musicChannel); }

        private var _songList:Array;
        private var internalSongList:Array; // song list from which we have cull any broken links
        /**
         * List of songs that MusicUtil will play from
         */
        public function get songList():Array { return _songList; }


        /**
         * List of songs that MusicUtil will play from
         */
        public function set songList(value:Array):void
        {
            if (null != value)
            {
                _songList = value.concat();
                internalSongList = value.concat();
            }
            else
            {
                _songList = [];
                internalSongList = [];
            }
        }

        /**
         * Get the MusicUtil ready for use.  Note that if you don't provide
         * any songs here, you better do a .SongList = [...];
         * @param	playlist and array of stego.util.music.Songs
         */
        public function init(playlist:/*stego.util.music.Song*/Array = null):void
        {
            songList = playlist;
        }
        /**
         * Clean up anything this could possibly leave hanging
         */
        public function shutdown():void
        {
            stopMusic();
            cleanUpSong();
        }

        /**
         * The song is going away or turned out to be invalid or something.
         * lose references to it
         */
        private function cleanUpSong():void
        {
            if (null != music)
            {
                try
                {
                    music.close();
                }
                catch (e:IOError)
                {
                }

                music.removeEventListener(IOErrorEvent.IO_ERROR, soundFail);
                music.removeEventListener(Event.OPEN, soundLoadBegin);
                music = null;
            }

            if (null != _currentSong)
            {
                _currentSong = null;
            }
        }

        /**
         * Start looping a random song.
         */
        public function startNewMusic():void
        {
            cleanUpSong();

            if (internalSongList.length > 0)
            {
                music = new Sound();
                _currentSong = ArrayUtil.randomOneOf(internalSongList);
                var file:String = currentSong.url;
                var request:URLRequest = new URLRequest(file);
                var context:SoundLoaderContext = new SoundLoaderContext(8000, false);

                music.addEventListener(IOErrorEvent.IO_ERROR, soundFail);
                music.addEventListener(Event.OPEN, soundLoadBegin);
                music.load(request, context);
            }
        }

        /**
         * music has been successfully found and can start playing
         */
        private function soundLoadBegin(e:Event):void
        {
            dispatchEvent(new MusicEvent(MusicEvent.EVENT_NEW_SONG, currentSong));
            beginPlaying();
        }

        /**
         *
         */
        private function soundFail(e:IOErrorEvent):void
        {
            ArrayUtil.removeEqual(internalSongList, _currentSong);
            startNewMusic();
        }

        /**
         * wrapper around StartNewMusic that can be used as an event listener
         * @param	e
         */
        private function _startNewMusic(e:Event = null):void
        {
            startNewMusic();
        }

        private function beginPlaying(e:Event = null, position:Number = 0):void
        {
            if (null != music)
            {
                stopMusic();

                _songVolume = currentSong.volume;
                musicChannel = music.play(position, 1, new SoundTransform(_volume * _songVolume) );
                musicChannel.addEventListener(Event.SOUND_COMPLETE, _startNewMusic);
            }
        }

        /**
         * Stop the current song
         */
        public function stopMusic():void
        {
            if (null != musicChannel)
            {
                musicChannel.removeEventListener(Event.SOUND_COMPLETE, _startNewMusic);
                musicChannel.stop();
                musicChannel = null;
            }

            pausePosition = 0;
        }

        /**
         * Pause the current song
         */
        public function pause():void
        {
            if (null != musicChannel)
            {
                pausePosition = musicChannel.position;
                musicChannel.stop();
                musicChannel = null;
            }
        }

        /**
         * Resume after pausing
         */
        public function resume():void
        {
            beginPlaying(null, pausePosition);
        }
    }

}
